/*
 * Fred's device driver for mices and keyboards V1.0
 * (C) Fred Hucht 1995
 * EMail: fred@thp.Uni-Duisburg.DE
 * Distributed in terms of the GNU public license
 */

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/fdb.h>
#include "fred.h"

/*
 * IMPORTANT: The following 3 addresses are patched to the real 
 * values in init (installed as /etc/install.d/init.d/fred)
 */
static void (**fred_int)() = (void(**)()) NULL;
					/* pointer to fdb interrupt table.
					 * 0x11008794 under 3.0.2 */

static Uchar *fred_id = (Uchar*) NULL;	/* pointer to fdb id table. 
					 * 0x12001DA8 under 3.0.2 */

static struct adbmsg *fred_automsg = (struct adbmsg*) NULL;
					/* pointer to fdb message area.
					 * 0x12001DF4 under 3.0.2 */

static int fred_mouse_state = 0x7;	/* actual state of the mouse buttons */
static int fred_debug = 0;		/* debug? */
static int fred_ok    = 1;		/* Everything OK? */
static void (*orig_mouse_intr)();	/* original mouse interrupt handler */

static void fred_mouse_intr(id, what, data)
int id, what;
unsigned short data;
{
  int spl = splimp(); 			/* Hope this is enough to mask interrupts */
  static int old_mouse_state = 0x7; 	/* all up */
  static unsigned short fred_button = 0x8000;
  Uchar b1, b2, b3;
  
  if(what == FDB_POLL) {
    b1 = (fred_automsg->data[0] & 0x80) >> 5;
    b2 = (fred_automsg->data[1] & 0x80) >> 6;
    b3 = (fred_automsg->data[2] & 0x80) >> 7;
    
    /* code mouse button state in fred_mouse_state for later query via ioctl:
     * left   button: bit 2 (4)
     * middle button: bit 1 (2)
     * right  button: bit 0 (1) */
    fred_mouse_state = b1 + b2 + ((fred_automsg->count > 2) ? b3 : 1);
    
    if(fred_mouse_state != old_mouse_state) {
      /* mouse state changed, report something */
      fred_button ^= 0x8000; /* flip button bit */
      old_mouse_state = fred_mouse_state;
    }
    data = (data & 0x7FFF) | fred_button;
  }
  
  if(fred_debug)
    printf("fred_intr: 0x%x 0x%x 0x%x 0x%x msg={0x%x 0x%x 0x%x {0x%x 0x%x 0x%x}} 0x%x\n",
	   id, what, data, fred_button,
	   fred_automsg->flags,   fred_automsg->count,
	   fred_automsg->cmd,     fred_automsg->data[0],
	   fred_automsg->data[1], fred_automsg->data[2],
	   fred_mouse_state);
  
  splx(spl);
  /* call normal mouse handler */
  /* if(id == 0) */ orig_mouse_intr(id, what, data);
}

void fred_init() {
  printf("Fred's FDB hacks (c) 1995 Fred Hucht. EMail: fred@thp.Uni-Duisburg.DE\n");
  
  if(fred_int     == NULL ||		/* This should not happen! */
     fred_id      == NULL ||		/* Driver is not installed as it should */
     fred_automsg == NULL) {
    printf("fred_init: Fatal! Driver not available.\n");
    fred_ok = 0;
  } else {
    fred_ok = 1;
    orig_mouse_intr = fred_int[FDB_MOUSE];    /* Save original mouse interrupt handler */
  }
}

int fred_open(dev, flag, ndevp)
dev_t	dev;
int	flag;
dev_t	*ndevp;
{
  if(fred_debug) {
    printf("fred_open: Called.\n fdb_int[0-3] = 0x%x 0x%x 0x%x 0x%x\n fdb_id[0-3] = 0x%x 0x%x 0x%x 0x%x\n",
	   fred_int[0], fred_int[1], fred_int[2], fred_int[3],
	   fred_id[0],  fred_id[1],  fred_id[2],  fred_id[3]);
#if 0
    {
      int i, j;
      printf("mouse_call: "); 
      for(i = 0; i<16; i++) { printf("%x ", *((Uchar*)0x12001f6c + i)); }
      printf("\n");
      printf("fdb_timeout: "); 
      for(i = 0; i < 64; i++) { printf("%x ", *((Uchar*)0x12028ba0 + i)); }
      printf("\n");
      printf("fdb_polling: ");
      for(i = 0; i < 32; i++) {
	for(j = 0; j < 32; j++) {
	  printf("%x ", *((Uchar*)0x12001ce4 + 32*i + j));
	}
	printf("\n");
      }
    }
#endif
  }
  return(0);
}

int fred_close(dev, flag)
dev_t	dev;
int	flag;
{
  if(fred_debug) printf("fred_close: Called.\n");
  return(0);
}

#define TRYMAX 1

int fred_ioctl(dev, cmd, addr, mode)
dev_t   dev;
int     cmd;
caddr_t addr;
int     mode;
{
  int ret = -1, try;
  struct fredmsg *msg;
  int h_id;
  Uchar data[8];
  
  if(fred_debug) 
    printf("fred_ioctl: dev=0x%x cmd=0x%x *addr={0x%x 0x%x 0x%x 0x%x ...} mode=0x%x\n",
	   dev, cmd, 
	   ((Uchar*)addr)[0],
	   ((Uchar*)addr)[1],
	   ((Uchar*)addr)[2],
	   ((Uchar*)addr)[3],
	   mode);
  
  switch (cmd) {
  case FRED_LISTEN:
    msg = (struct fredmsg *)addr;
    if(msg->addr & 0xFFFFFFF0) /* valid FDB address? */
      return(EINVAL);
    if(fred_int[msg->addr] == NULL) /* device open? */
      return(EINVAL);
    
    for(try = 0; try < TRYMAX && ret != 0; try++) {
      ret = fdb_listen(msg->addr,
		       msg->id, /* fred_id[msg->addr] */
		       msg->reg,
		       msg->data,
		       msg->count);
      if(fred_debug) printf("fred_ioctl LISTEN: ret = %d\n", ret);
    }
    
    msg->ret = ret;
    return(0); 
    break;
  case FRED_TALK:
    msg = (struct fredmsg *)addr;
    if(msg->addr & 0xFFFFFFF0) /* valid FDB address? */
      return(EINVAL);
    if(fred_int[msg->addr] == NULL) /* device open? */
      return(EINVAL);
    
    for(try = 0; try < TRYMAX && ret != 0; try++ ) {
      msg->ret = fdb_talk(msg->addr,
			  msg->id, /* fred_id[msg->addr] */
			  msg->reg,
			  msg->data);
      if(fred_debug) printf("fred_ioctl TALK: ret = %d\n", ret);
    }
    return(0); 
    break;
  case FRED_MOUSE:
    /* returns mouse button state :
     * bit 4 : left   button up
     * bit 2 : middle button up
     * bit 1 : right  button up
     */
    *((int*)addr) = fred_mouse_state;
    return(0); 
    break;
  case FRED_DEBUG:
    /* enables/disables debug info to /dev/console */
    fred_debug = *((int*)addr); 
    return(0); 
    break;
  case FRED_SETKID:			/* set keyboard device handler ID */
    h_id = *(int*)addr;
    if(h_id < 0 || h_id > 255) return(EINVAL);
    
    data[0] = FDB_KEYBOARD;		/* don't change highbyte */
    data[1] = h_id;			/* set device handler ID */
    
    for(try = 0; try < TRYMAX && ret != 0; try++ ) {
      ret = fdb_listen(FDB_KEYBOARD,	/* FDB address */
		       0x0,		/* FDB id */
		       0x3,		/* register */
		       data,		/* handler ID */
		       2);		/* count */
      
      if(fred_debug) printf("fred_ioctl SETKID: ret = %d\n", ret);
    }
    
    *(int*)addr = ret;
    if(ret != 0) return(EBUSY);
    else         return(0); 
    break;
  case FRED_SETMID:			/* set mouse device handler ID */
    h_id = *(int*)addr;
    if(h_id < 0 || h_id > 255) return(EINVAL);
    
    data[0] = FDB_MOUSE;		/* don't change highbyte */
    data[1] = h_id;			/* set device handler ID */
    
    for(try = 0; try < TRYMAX && ret != 0; try++ ) {
      ret = fdb_listen(FDB_MOUSE,	/* FDB address */
		       0x0,		/* FDB id */
		       0x3,		/* register */
		       data,		/* handler ID */
		       2);		/* count */
      
      if(fred_debug) printf("fred_ioctl SETMID: ret = %d\n", ret);
    }
    
    *(int*)addr = ret;
    if(ret != 0) return(EBUSY);
    else         return(0);
    break;
  case FRED_INSTM:
    if(fred_debug) printf("fred_ioctl: INSTM.\n");
    if(orig_mouse_intr == NULL)	{	/* Strange: No mouse interrupt */
      return(ENODEV);			/* handler installed */
    } else {
      fred_int[FDB_MOUSE] = fred_mouse_intr;    /* install our interrupt handler */
      return(0); 
    }
    break;
  case FRED_UNINSTM:
    if(fred_debug) printf("fred_ioctl: UNINSTM.\n");
    if(orig_mouse_intr == NULL)	{	/* Strange: No mouse interrupt */
      return(ENODEV);			/* handler installed */
    } else {
      fred_int[FDB_MOUSE] = orig_mouse_intr;    /* install default interrupt handler */
      return(0); 
    }
    break;
  default:
    return(EINVAL);
    break;
  }
  return(EINVAL);			/* Should not happen */
}

int fred_write() { return(EINVAL); }
int fred_read()  { return(EINVAL); }
